
package require direct 0.3


namespace eval meta {
	set name "ICQ Direct"
	set description "ICQ Direct connections"
	set author "Ihar Viarheichyk <iverg@mail.ru>"

	proc unload {} {
		[namespace parent]::stop
		# Close all pending connections
		upvar 0 [namespace parent]::dc dc
		if {[info exists dc]} {
			foreach {key val} [array get dc] { $val delete }
			unset dc
		}
	}
}

# Start DC server
proc start {} {
	variable dcserver
	if {[info exists dcserver]} return
	if {![my is ICQ-Account]} { 
		Event Log error "ICQ module is not loaded, DC is not available"
	} else {
		set dcserver [direct::server 0 -uin\
			[set [module::space icq]::uin] -event [nc DC-Event]]
		[my icq-command] configure -dcinfo [$dcserver info]
		my log info "Started DC server"
	}
}

# Stop DC server
proc stop {} {
	variable dcserver
	if {[info exists dcserver]} {
		$dcserver delete
		unset dcserver
		[my icq-command] configure -dcinfo ""
		my log info "Stopped DC server"
	}
}

method ICQ-DC-Account set-status {status} {
	variable dcserver
	set exist [info exists dcserver]
	expr {($status == "offline")?[stop]:[start]}
}

namespace eval contacts {
	include ICQ [namespace current] 0.45

	# Show Direct Connection IP in cntact node hint
	proc hint {} {
		upvar 1 self self
		upvar 1 result result
		variable ips

		if {[info exists ips($self)] && [info exists result]} {
			return "$result\nDC IP: $ips($self)"
		} else { return -code continue }
	}

	# We are only interested if client unicode-aware or not
	proc client-info {client} {
		upvar 1 self self
		if {[lindex $client end]} {
			upvar 0 [namespace parent]::dc($self) dc
			if {[info exists dc]} {
				$dc configure -encoding utf-8
			}
		}
	}
}

# Create Direct connection handler for contact
proc contacts::dc-info {type ip port cookie version} {
	upvar 1 self self
	upvar 0 [namespace parent]::dc($self) dc
	variable ips
	set uin [lindex $self end]
	set ref [$self property]
	
	# If DC is disabled, delete previous instance, if present
	if {$type=="disabled" || $type=="web" || $ip=="0.0.0.0" || $port==0} {
		if {[info exists dc]} { 
			$dc delete
			unset dc 
			unset ips($self)
		}
		return
	}
	set ips($self) $ip
	# If no previous instance found, create new DC handler
	if {![info exists dc]} { 
		set dc [direct::peer $uin -event [namespace parent]::DC-Event]
	}
	upvar 0 [module::space icq]::encoding encoding
	if {[info exists ${ref}(encoding)]} {
		set enc [set ${ref}(encoding)]
	} else { set enc $encoding }

	variable dcserver
	if {[info exists dcserver]} {
		set mp [lindex [$dcserver info] 0]
	} else { set mp 0 }
	$dc configure -host $ip -port $port -cookie $cookie\
		-uin [set [module::space icq]::uin] -encoding $enc -myport $mp
	my log info "DC connection to $uin [$dc state]"
	my log warning "DC cookie for $uin $cookie"
}

# Send message directly
method ICQ-DC send {type message} {
	upvar 1 self self
	variable dc
	# Send only text, URL and contacts via DC
	if {[lsearch -exact {text URL contacts} $type]==-1} {
		return -code continue
	}
	$dc($self) send $type $message
}

# Handle DC events
proc DC-Event {uin event args} {
	if {[info commands hooks::$event]!=""} {
		eval [list hooks::$event [list ICQ $uin]] $args
	}
}

# Direct connection command events '''1
namespace eval hooks {
	proc Established {id} {
		my log {info DC} "Connection to $id established"
		$id include ICQ-DC 0.25
	}

	proc Disconnected {id reason} {
		# If connection was closed by peer, reconnect
		if {[$id is ICQ-DC] && [$id status]!="offline"} {
			upvar 0 [namespace parent]::dc($id) dc
			$id exclude ICQ-DC
			$dc connect
		}
	}
	
	# Message acknowledge
	proc ACK {id type msgid} { $id acknowledged $msgid $type }

	proc Log {id tags str} { 
		Event Log [concat $tags ICQ:DC] "Peer $id: $str"
	}

	# Handle incoming message
	proc Incoming {id type message} {
		$id incoming $type $message [clock seconds]\
			[clock seconds]:[clock clicks] 
	}

	# Timeout, resend via server
	proc Timeout {id type message msgid} {
		Event Log notice "DC timoeout, resending via server"
		[my icq-command] send $type [lindex $id end]\
			$message $id
	}

}
# End of DC hooks ```1

# Initialize DC on chat open '''1
namespace eval init {
	include Chat [namespace current] 0.7

	proc initialize {args} {
		upvar 1 self self
		set id [lindex $self end]
		if {[$id is ICQ] && ![$id is ICQ-DC]} {
			upvar 0 [namespace parent]::dc($id) dc
			if {[info exists dc]} { $dc connect }
		}
	}

	proc destroy {} {
		upvar 1 self self
		set id [lindex $self end]
		if {[$id is ICQ-DC]} {
			upvar 0 [namespace parent]::dc($id) dc
			Event Log {info DC} "Close DC to $id"
			# Indicate that connection is closed by us
			$id exclude ICQ-DC
			$dc disconnect
		}
	}
}
# ```1

include ICQ-Account ICQ-DC-Account 0.6

if {[my status]!="offline"} start

