# This module is a part of Alicq instant messenger package
# It provides core of basic graphical interface of Alicq: tree window
#
# Author: Ihar Viarheichyk

# Contact list interface - it .tr widget exists, another contact list
# module is loaded
if {[winfo exists .tr]} {
	return -code "Contact list is already initialized, skipping"
}

package require BWidget
package require resource

event add <<ContextCurrent>> <3> <Control-1>
event add <<ContextSelected>> <F9>
event add <<Item-Activate>> <Double-1> <Return>

option add *status.compound left

font create group -family [font actual normal -family]\
	-size [font actual normal -size] -weight bold

namespace eval meta {
	variable author "Ihar Viarheichyk <iverg@mail.ru>"
	variable name "Contact list"
	variable description "Basic group and contact list implementation"
	variable icon img:group

	array set open { type cache save change default "" }
	array set order { type cache save change default "" }

	proc unloadable {} { return 0}
}

# Node hints '''1
namespace eval hint {
	proc init {tree} {
		foreach id {Text Image} {
			$tree bind${id} <Enter> [nc show]
			$tree bind${id} <Leave> [nc close]
		}
	}

	proc show {node} {
		ui::tooltip::show [list [.tr itemcget $node -data] hint]
	}

	proc close {args} { ui::tooltip::close }
}
# End of node hints ```1

# Tree widget handling routines '''1

proc tree {} {
	set sw [ScrolledWindow .sw -relief flat -borderwidth 0]
	set tree [Tree .tr -deltay [DeltaY]\
		-dragenabled yes -dropenabled yes -dropcmd [nc Dropper]]
	$sw setwidget $tree
	foreach id {Text Image} {
		$tree bind${id} <Double-1> [nc DoubleClick]
		$tree bind${id} <<ContextCurrent>> [nc menu::Popup %X %Y]
	}

	# Small hack to allow mouse wheel scrolling in tree widget
	bind .tr.c <4> ".tr.c yview scroll -2 units"
	bind .tr.c <5> ".tr.c yview scroll 2 units"

	bind .tr.c <<ContextSelected>> [nc menu::PopupSelected %X %Y]
	bind . <Return> [namespace code  {DoubleClick [.tr selection get]}]
	bind [winfo class .] <FocusIn> { focus .tr.c }
	set sw
}

proc DeltaY {} {
	set max 0
	foreach x [image names] {
		set h [image height $x]
		if {[string match img:* $x] && $h>$max} {set max $h}
	}
	incr max 2
}

# Handle double click on tree node
proc DoubleClick {node} {
	if {[llength [.tr nodes $node]]} {
		.tr itemconfigure $node -open\
			[expr [.tr itemcget $node -open]^1]
	} else { [.tr itemcget $node -data] popup }
}

# --- Handle in-tree item renaming
method TreeItem::menu {.98 Rename} {} {
	upvar 1 self self
	variable CurrentNode
	.tr edit $CurrentNode [$self alias] [nc ItemModify $self]
}

proc ItemModify {id text} {
	$id new-alias $text
	return 1
}

# New-style drop handler
proc Dropper {dest src type op datatype data} {
	if {$datatype!="TREE_NODE"} return
	if {[lindex $type 0]!="node"} {
		puts "Type $type, src $src, dest $dest, op $op, datatype $datatype, data $data"
		return
	}
	set to [lindex $type 1]

	set dest_id [.tr itemcget $to -data]
	# Do not handle dropping on itself or own parent, 
	# and destination item is a group
	if {$data==$to || [.tr parent $data]==$to || 
	    ![$dest_id is Group] } return

	# Check if the object is already exists in destination
	set id [.tr itemcget $data -data]
	foreach x [.tr nodes $to] {
		if {[.tr itemcget $x -data]==$id} return
	}

	# Allow moving only to user groups, otherwise make copy
	if {![$dest_id is UserGroup]} { set op copy }
	
	# Handle operation
	switch -- $op {
		copy { $dest_id add $id }
		move -
		default { 
			$dest_id add $id
			[.tr itemcget [.tr parent $data] -data] remove $id
		}
	}
}

# --- Handle deletion of an item
method TreeItem::menu {0.99 Delete} {} {
	upvar 1 self self
	variable CurrentNode

	if {[lindex $self 0]=="group" && [llength [.tr nodes $CurrentNode]]} {
		tk_messageBox -type ok -title Error -message\
			"Can not remove [$self name] because it is not empty"
		return	
	}
	# If the only node remains in the tree, ask user to confirm 
	# delete item from the roster, otherwise delete from the group
	if {[llength $tree::mapping($self)]==1} {
		set name [$self alias]
		if {[tk_messageBox -type yesno -title "Removing $name"\
			-message "Remove $name from contact list?"]
					=="yes"} { $self destroy }
	} else { [.tr itemcget [.tr parent $CurrentNode] -data] remove $self }
}
# End of tree widget handling routines ```1

# Handle tree nodes mapping changes '''1
namespace eval tree {
	variable counter 0
	namespace export *

	# Show nodes for given object recursively
	proc show {id} {
		if {[info commands $id]==""} return
		variable mapping
		if {[info exists mapping($id)]} { return $mapping($id) }
		foreach x [$id parents] { show-helper $x $id }
		open::restore $id
		set mapping($id)
	}

	# Update existing nodes attributes. If obect is not mapped, do nothing
	proc update {id args} {
		variable mapping
		if {![info exists mapping($id)]} return
		foreach x $mapping($id) { 
			eval [list .tr itemconfigure $x] $args
		}
	}

	# Helper function to add nodes to the tree and resort parent nodes
	proc show-helper {x id} {
		variable counter
		variable mapping
		set txt [counters::text $id]
		set img [$id image]
		set res [list]
		foreach node [show $x] {
			set new [incr counter]
			.tr insert end $node $new -data $id\
				-text $txt -image $img
			lappend mapping($id) $new
			lappend res $new
		}
		reorder $x
		set res
	}

	# Hide nodes recursively, if "hide empty groups" flag is set
	proc hide {id} {
		variable mapping
		if {![info exists mapping($id)]} return
		open::save $id
		foreach x $mapping($id) { rechide $x }
		unset mapping($id)
	}

	# Hide nodes recursively (if hide empty flag is set)
	proc rechide {node} {
		if {![.tr exists $node]} return
		upvar 0 [namespace parent]::hide::empty hide
		set n [.tr parent $node]
		if {$hide} {
			if {[llength [.tr nodes $n]]==1 && $n!="root"} {
				hide [.tr itemcget $n -data]
				return
			} 
		}
		.tr delete $node
	}

	# This procedure updates mapping of all children of the node woithout
	# actual removal of tree item
	proc virtual-remove {n} {
		foreach x [.tr nodes $n] { virtual-remove $x }
		variable mapping
		set id [.tr itemcget $n -data]
		if {[info exists mapping($id)]} {
			if {[llength mapping($id)]==1} {
				unset mapping($id)
			} elseif {[set p [lsearch $mapping($id) $n]]!=-1} {
				set mapping($id) [lreplace $mapping($id) $p $p]
			}
		}
	}

	# Copy node with its children to new parent, updating
	# mapping properly
	proc copy {node parent} {
		variable counter
		variable mapping
		set new [incr counter]
		.tr insert end $parent $new
		foreach x [.tr itemconfigure $node] {
			.tr itemconfigure $new [lindex $x 0] [lindex $x end]
		}
		lappend mapping([.tr itemcget $node -data]) $new
		foreach x [.tr nodes $node] { copy $x $new }
	}

	proc reorder {n} {
		variable sorting
		if {![info exists sorting($n)]} {
			set sorting($n) [after idle [nc _reorder $n]]
		}
	}

	proc _reorder {n} {
		variable sorting
		variable sorter
		variable mapping
		unset sorting($n)
		if {![info exists mapping($n)]} return
		if {[info exists sorter($n)]} {
			 set cmd $sorter($n)
		} else { set cmd Status }	
		foreach x $mapping($n) { 
			.tr reorder $x [lsort -command [list cmp sort::$cmd]\
				[.tr nodes $x]]
		}
	}	

	proc cmp {cmd n1 n2} { 
		foreach x {1 2} { set id$x [.tr itemcget [set n$x] -data] }
		$cmd $id1 $id2
	}

	# Intercept 'add' method to update tree
	method TreeItem add {id} {
		variable mapping
		upvar 1 self self
		# If id has mapping, check if it has children to copy
		if {[info exists mapping($id)] && [llength $mapping($id)]} {
			set children [.tr nodes [lindex $mapping($id) 0]]
		} else {set children [list]}
		
		# Show node recursively
		set nodes [show-helper $self $id]
		# Copy children if any
		foreach x $children {
			foreach parent $nodes { copy $x $parent }
		}
	}

	# Intercept 'remove' method to update tree
	method TreeItem remove {id} {
		upvar 1 self self
		variable mapping
		if {![info exists mapping($self)]} return
		foreach x $mapping($id) {
			if {[.tr itemcget [.tr parent $x] -data] == $self} {
				# Update children mapping
				virtual-remove $x
				# Remove node
				rechide $x
			}
		}
		return -code continue
	}

	# Intercept TreeItem deletion method to remove node from the tree
	method TreeItem destroy {} {
		upvar 1 self self
		hide $self
		return -code continue
	}


	# Intercept new-status method to update node images
	method TreeItem new-status {status} {
		upvar 1 self self
		upvar 0 [namespace parent]::hide::offline hide
		if {$status=="offline" && $hide} { hide $self } else {
			foreach node [show $self] {
				.tr itemconfigure $node -image img:$status
				reorder [.tr itemcget [.tr parent $node] -data]
			}
		}
		# Check status transition and update online counter if needed
		counters::online-offline $self $status
		return -code continue
	}

	# On tree item alias change update proper nodes and reorder groups
	method TreeItem new-alias {value} {
		upvar 1 self self
		variable mapping
		if {[info exists mapping($self)]} {
			update $self -text [counters::text $self]
			foreach x $mapping($self) { 
				reorder [.tr itemcget [.tr parent $x] -data]
			}
		}
	}

	# Resisively open parent groups and make any of object nodes visible
	proc see {id} {
		variable mapping
		foreach node [show $id] { if {[.tr visible $node]} break }
		if {![.tr visible $node]} { recursively-open $node }
		.tr see $node
		.tr selection set $node
	}

	proc recursively-open {node} {
		if {$node=="root"} return
		.tr itemconfigure $node -open 1
		recursively-open [.tr parent $node]
	}

	proc init {} {
		variable mapping
		upvar 0 [namespace parent]::hide::offline offline
		upvar 0 [namespace parent]::hide::empty empty
		array set mapping {{tree root} root}
		
		# Restore sorting information
		variable sorter
		array set sorter [set [namespace parent]::order]

		counters::init
		if {!$empty} {
			foreach id [tags::select Group] { show $id }
		}
		#TODO: this condition is true for initial load
		# If loaded during operation, need to check contact status
		if {!$offline} {
			foreach id [tags::select Contact] { show $id }
		}
		
		open::restore_all

		# Intercept initialize method for new roster items
		# Assume new group is empty and new contact is offline
		method TreeItem initialize {args} {
			upvar 1 self self
			upvar 0 [namespace parent]::hide::empty empty
			upvar 0 [namespace parent]::hide::empty offline
			if {[$self is Group] && $empty ||
			    [$self is Contact] && $offline} return

			after idle [nc show $self]
		}

	}

	after idle [nc init]
}
# End of tree mapping changes ```1

# Save/restore group openness and ordering '''1
namespace eval tree::open {
	# Save group openness information
	proc save {id} {
		upvar 0 [namespace parent]::mapping mapping
		variable op
		if {![info exists mapping($id)]} return
		set i 0
		foreach node $mapping($id) {
			if {[.tr itemcget $node -open]} { lappend op($id) $i }
			incr i
		}
	}

	# Restore group openness information
	proc restore {id} {
		upvar 0 [namespace parent]::mapping mapping
		variable op
		if {![info exists op($id)] ||
		    ![info exists mapping($id)]} return
		foreach idx $op($id) {
			set node [lindex $mapping($id) $idx]
			if {$node==""} continue
			.tr itemconfigure $node -open 1
		}
		unset op($id)
	}

	# Save openness and sorting information for all groups
	proc save_all {} {
		variable op
		upvar 0 [namespace parent]::sorter sr
		foreach x [tags::select Group] { save $x }
		set [namespace parent [namespace parent]]::open [array get op]
		set [namespace parent [namespace parent]]::order [array get sr]
	}

	# Restore openness information for all groups
	proc restore_all {} {
		variable op
		array set op [set [namespace parent [namespace parent]]::open]
		foreach x [array names op] { restore $x }
	}
}
# End of group openness handling ```1

# This namespace contains functions related to group counters '''1
namespace eval tree::counters {
	variable contacts
	variable online

	# Initialize group contact counters
	proc init {} {
		variable contacts
		foreach x [tags::select Contact] {
			foreach p [$x parents] {
				if {[info exists contacts($p)]} {
					incr contacts($p)
				} else { set contacts($p) 1}
			}
		}
	}

	# Construct node text of object alias and counters, if available
	proc text {id} {
		variable contacts
		variable online
		set txt [$id alias]
		if {[info exists contacts($id)]} { append txt " - "\
			[expr {[info exists online($id)]?$online($id):0}]\
			/ $contacts($id)}
		set txt
	}

	# Handle online-offline transition to update parent counters
	proc online-offline {self status} {
		variable last
		if {$status=="offline" && [info exists last($self)]} {
			set c -1
			unset last($self)
		} elseif {![info exists last($self)] && $status!="offline"} {
			set c 1
			set last($self) 1
		}
		if {[info exists c]} {
			variable online
			foreach x [$self parents] {
				if {![info exists online($x)]} {
					set online($x) $c
				} else { incr online($x) $c }
				[namespace parent]::update $x -text [text $x]
			}
		}
	}	
}
# End of counters handling ```1

# Toolbar triggers. '''1
# Currently include show/hide offline contacts and
# show/hide empty groups triggers
namespace eval hide {

	# Default toolbar panel resources
	option add *quick.Checkbutton.selectColor "" widgetDefault
	option add *quick.Checkbutton.indicatorOn 0
	option add *quickPanel true widgetDefault

	# Default keyboard bindings
	event add <<Hide-Offline>> <Control-o>
	event add <<Hide-Empty>> <Control-i>

	namespace eval meta {
		array set empty {
			type boolean default 0
			description "Hide empty groups" save change
			
		}
		array set offline {
			type boolean default 0
			property Global:UI:Tree|HideOffline
			description "Hide offline contacts" save change
		}
	}
	
	# Add main menu items for triggers
	method Me::menu [list Interface "Hide offline contacts"] {args} {
		eval [list set [namespace current]::offline] $args
	}

	method Me::menu [list Interface "Hide empty groups"] {args} {
		eval [list set [namespace current]::empty] $args
	}

	proc init {} {
		# Initialize triggers
		foreach x {offline empty} {
			variable $x
			trace variable $x w [nc trigger_$x]
			bind . <<Hide-[string totitle $x]>> [nc trigger $x]
		}
	}
	
	# Initialize toolbar panel
	proc panel {top} {
		frame $top
		if {[option get . quickPanel Boolean]=="true"} {
			grid $top -columnspan 2 -sticky we
		}
		grid columnconfigure $top 8 -weight 1
		grid [checkbutton $top.offline -image img:offline\
			-variable [namespace current]::offline]\
			-row 0 -column 1 -padx 2 -ipadx 2 -ipady 2 -sticky w
		ui::tooltip::action $top.offline\
			[mc $meta::offline(description)] <<Hide-Offline>>
		grid [checkbutton $top.empty -image img:group\
			-variable [namespace current]::empty]\
			-row 0 -column 2 -padx 2 -ipadx 2 -ipady 2 -sticky w
		ui::tooltip::action $top.empty\
			[mc $meta::empty(description)] <<Hide-Empty>>
	}

	proc trigger {name} {
		variable $name 
		set $name [expr [string is true [set $name]]^1] 
	}

	# handle hide offline flag trigger
	proc trigger_offline {args} {
		variable offline
		set cmd [expr $offline?"hide":"show"]
		foreach x [tags::select Contact] {
			if {[$x status]=="offline" && ![$x is Pending]} { 
				$cmd $x 
			}
		}
	}

	# Handle empty groups show/hide trigger
	proc trigger_empty {args} {
		variable empty
		empty-[expr {$empty?"hide":"show"}]
	}

	proc empty-show {} { foreach id [tags::select Group] { show $id } }

	proc empty-hide {} {
		upvar 0 [namespace parent]::tree::mapping mapping
		foreach id [tags::select Group] {
			if {![info exists mapping($id)]} continue
			foreach x $mapping($id) {
				if {![llength [.tr nodes $x]]} { hide $id }
				break
			}
		}
	}

	namespace import [namespace parent]::tree::show
	namespace import [namespace parent]::tree::hide
	after idle [nc init]
}
# End of triggers ```1

# Flashing icon when pending messages present '''1
namespace eval Pending {
	foreach x {hide show update see} {
		namespace import [namespace parent]::tree::$x
	}	

	proc Flasher {stage} {
		variable scheduled
		foreach x [tags::select Pending] { flash $x $stage }
		set scheduled [after 300 [nc Flasher [expr $stage^1]]]
	}

	proc flash {id stage} {
		update $id -image img:[expr {$stage?[$id status]:"message"}]
	}	

	method TreeItem pending {args} {
		variable scheduled
		upvar 1 self self
		see $self
		if {![info exists scheduled]} { Flasher 0 }
	}

	method TreeItem no-pending {args} {
		upvar 1 self self
		flash $self 1
		if {[set [namespace parent]::hide::offline] &&
			 [$self status]=="offline"} { hide $self }
		if {[llength [tags::select Pending]]<2} {
			variable scheduled
			after cancel $scheduled
			unset scheduled
		}
		after idle [nc next]
	}

	# Find contact havind pending message and make it viewable in the tree
	proc next {} { foreach x [tags::select Pending] { see $x; break } }
}
# End of pending messages handling ```1

# Contact list quick search fucntions '''1
namespace eval QuickSearch {
	event add <<QuickSearch>> <Key-slash> <Control-Key-f>
	event add <<SearchNext>> <Key-F3>
	event add <<SearchCancel>> <FocusOut> <Key-Escape>

	namespace import [namespace parent]::tree::see
	namespace import [namespace parent]::tree::show
	namespace import [namespace parent]::tree::hide

	# Add invoke QuickSearch from menu
	method Me::menu {0.099 Search {In Contact List}} {} {
		set [namespace parent]::quick_search 1
	}

	# Create quick search widget
	proc init {} {
		set searchVar [namespace current]::quick_search
		grid [checkbutton .quick.search -text [mc Search]\
			-variable $searchVar] -row 0 -column 10\
			-padx 2 -ipadx 2 -ipady 2 -sticky e
		ui::tooltip::action .quick.search\
			[mc "Search in contact list"] <<QuickSearch>>
		set top .qsearch
		trace variable $searchVar w [nc toggle $top.f .tr.c]
		bind .tr.c <<QuickSearch>> [list set $searchVar 1]

		# Create search widget
		frame $top
		entry $top.f -validate key -vcmd [nc find $top.f %P]
		button $top.next -text [mc Next] -command [nc next]\
			-state disabled
		ui::tooltip::action $top.next\
			[mc "Next matching contact"] <<SearchNext>>
		grid $top.f $top.next -sticky news -pady 2 -padx 2
		grid columnconfigure $top 0 -weight 1
		bind $top.f <<SearchCancel>> [list set $searchVar 0]
		bind $top.f <Up> [nc history $top.f]
		bind $top.f <<SearchNext>> [list $top.next invoke]
	}

	# Handle search history
	proc history {name} {
		variable history
		variable last_pos
		if {![info exists history]} return
		set cur [$name get]
		set next [lindex $last_search end-$last_pos]
		if {$next==""} { set last_pos 0 } else { incr last_pos }
		$name delete 0 end
		$name insert end $next
	}

	# Update search history on search close
	proc update-history {val} {
		variable history
		# Find if result is already stored in history.
		if {[info exists history] && 
		    [set pos [lsearch -exact $history $val]] != -1 } {
			set history [lreplace $history $pos $pos]
		}
		lappend history $val
	}

	# Check if object matches given search string
	# Check alias or object ID if alias does not match
	proc matches {id val} {
		set len [string length $val]
		expr {[string equal -nocase -length $len $val [$id alias]] ||
		      [string equal -nocase -length $len $val [lindex $id end]]}
	}

	# Perform search
	proc find {entry val} {
		variable found
		variable index
		if {[info exists index]} { unset index }
		set last [expr {[info exists found]?$found:[list]}]
		# If empty value is given as matching pattern, hide items from
		# possible last search and return as 'valid'
		if {$val==""} { 
			set found [list]
			cond-hide $last
			[winfo parent $entry].next configure -state disabled
			return 1
		}

		# Find items matching search pattern
		set found [grep x [tags::select Contact] { [matches $x $val] } ]
		# Show items if needed
		foreach x $found { 
			if {[set pos [lsearch -exact $last $x]]!=-1} {
				set last [lreplace $last $pos $pos ]
			} else { show $x }
		}
		# Hide items from prevous search
		if {[llength $found]} { cond-hide $last }
		unset last
		# Move to the first item of result
		next
		[winfo parent $entry].next configure -state\
			[expr {([llength $found]>1)?"normal":"disabled"}]
		llength $found
	}

	# Hide items from last search if needed
	proc cond-hide {list} {
		upvar 0 [namespace parent]::hide::offline hide
		if {!$hide} return
		foreach x $list {
			if {[$x status]=="offline" &&
			    ![$x is Pending]} { hide $x }
		}
	}

	# Show next element of the list of matching items
	proc next {} {
		variable found
		if {![llength $found]} return

		variable index
		if {![info exists index]} { set index -1 }
		if {[llength $found]-$index < 2} { set index -1 }
		see [lindex $found [incr index]]
	}

	# Toggle Quick Search on or off
	proc toggle {name main var args} {
		upvar 1 $var mode
		set current [winfo viewable $name]
		if {$current==$mode} return 
		variable last_pos
		if {$mode} {
			grid [winfo parent $name]\
				-column 0 -columnspan 2 -row 4 -sticky we
			set last_pos 0
			focus $name
		} else { 
			# Hide results of the search if needed	
			variable found
			if {[info exists found]} {
				cond-hide $found
				unset found
			}
			# Remove result index as well
			variable index
			if {[info exists index]} { unset index }

			grid forget [winfo parent $name] 
			# Update history
			if {[set val [$name get]]!=""} { update-history $val }
			# Unset last history position
			if {[info exists last_pos]} { unset last_pos }
			# Delete search pattern
			$name delete 0 end
			focus $main
		}
	}
}
# End of Quick Search namespace ```1

# This namespace contains all available sorting methods '''1
namespace eval tree::sort {
	proc Status {id1 id2} {
		set w {"" ffc online occ dnd invisible away na offline}
		foreach x {1 2} {
			set w$x [lsearch -exact $w [[set id$x] status]]
		}
		return [expr {($w1<$w2)?-1:(($w1>$w2)?1:[Alphabet $id1 $id2])}]
	}

	proc Alphabet {id1 id2} {
		string compare -nocase [$id1 alias] [$id2 alias]
	}

	proc Identifier {id1 id2} { string compare $id1 $id2 }
}
# End of sorting methods ```1

# Handle account status and alias change '''1
namespace eval account {
	include Me [namespace current] 0.7

	proc new-status {status} {
		set status [my status]
		.status configure -image img:$status -text [mc $status]
		ui::tooltip::action .status\
		   "[mc {Change status}]\n[mc {Now status is}] [mc $status]"
		new-alias
	}

	# Update main window title on alias change. This procedure is also
	# called when status is changed
	proc new-alias {args} {
		wm title . "Alicq: [my alias] [my status]"
	}

	# On exit, generate openness information
	proc exit {args} {
		[namespace parent]::tree::open::save_all
	}

	# Initial window title
	after idle [nc new-alias]
}
# End of account-wide handling methods ```1

# Main and context menu handling '''1
namespace eval menu {

	# Create status menu and generate initial main menu
	proc init {} {

		# Check if we are running in Aqua environment
		set aqua [string equal\
			[catch { [tk windowingsystem]}] "aqua"]

		# Creare main menu
		if {$aqua} {
			menu .main -type menubar -tearoff no
			. configure -menu .main
			.main add cascade -menu .main.apple
			# Display node context menu as dropdown
			# of application menu
			.main add cascade -menu .main.context
			bind .tr <<TreeSelect>> [nc NodeSelected %W]
			proc NodeSelected {tree} {
				set nodes [$tree selection get]
				if {[llength $nodes]!=1} return
				set node [lindex $nodes 0]
				ui::context-menu .main.context\
					[$tree itemcget $node -data]
				.main entryconfigure last -label\
					[$tree itemcget $node -text]
				variable CurrentNode
				set CurrentNode $node
			}
		} else {
			set menu1 [menubutton .main -text [mc "Menu"]\
				-direction above -menu .main.apple]
			ui::tooltip::action .main [mc "Invoke main menu"]
			grid $menu1 -column 1 -row 5 -sticky nes -padx 2
		}
		menu .main.apple -type normal -tearoff no

		# Create status selection menus
		set menu2 [menubutton .status -text [mc offline]\
			-menu .status.m -image img:offline]
		menu .status.m -type normal -tearoff no
		grid $menu2 -column 0 -row 5 -sticky nws -padx 2 

		# Populate status selection menu
		foreach status {online away dnd occ ffc na invisible offline} {
			.status.m add command -label [mc $status]\
				-command [list my set-status $status]
			# Disbale status images for menu items in Aqua	
			if {!$aqua} { 
				.status.m entryconfigure last\
					-image img:$status 
			}
		}

		# Main menu content
		ui::context-menu .main.apple my

		# TODO: replace with new-style methods
		handler {ModuleLoaded ModuleUnloaded} recalcMenu {module} { 
			ui::context-menu .main.apple my 
		} 0.99
	}

	# --- Context menu handling
	proc PopupSelected {x y} {
		set sel [.tr selection get]
		if {[llength $sel]} { Popup $x $y [lindex $sel 0] }
	}

	proc Popup {x y node} {
		set [namespace parent]::CurrentNode $node
		ui::context-menu .context [.tr itemcget $node -data]
		tk_popup .context [expr $x-10] [expr $y-5]
	}

	after idle [nc init]
}

# Menu items
method Group::menu {"Order by..."} {args} {
	upvar 1 self self
	if {$args==""} { 
		if {[info exists tree::sorter($self)]} { 
			set val $tree::sorter($self)
		} else {set val Status }
		return [list $val [Available [namespace current]::tree::sort]]
	} else { 
		set tree::sorter($self) [lindex $args 0]
		tree::reorder $self
	}
}

method Group::menu {"Color mark"} {args} {
	upvar 1 self self
	if {$args==""} {
		if {[info exists tree::mark($self)]} {
			set val $tree::mark($self)
		} else { set val black }
		return [list $val {black blue red green brown yellow}]
	} else {
		set tree::mark($self) [lindex $args 0]
		foreach x $tree::mapping($self) {
			foreach node [.tr nodes $x] {
				.tr itemconfigure $node -fill [lindex $args 0]
			}
		}
	}
}

method TreeItem::menu Groups {} {
	upvar 1 self self
	ui::groups::top $self
}

# End of menu handling ```1

# Main tree window '''1
proc MainWindow {} {
	wm group . .
	wm command . [file join [pwd] $::argv0]
	# Handle geometry command-line option
	if {![info exists ::geometry] &&
	    [set gm [option get . geometry Geometry]]!=""} {wm geometry . $gm }
	# Handle state of main window on startup: normal or iconic	 
	if {[set state [option get . state State]]!=""} { wm state . $state }

	# Toolbar panel
	hide::panel .quick

	# Contact list widget
	grid [tree] -columnspan 2 -row 3 -sticky news

	# Initialize tooltips
	hint::init .tr

	# Quick Search
	QuickSearch::init

	grid columnconfigure . 1 -weight 1
	grid rowconfigure . 3 -weight 1

	if {[wm protocol . WM_DELETE_WINDOW]==""} {
		wm protocol . WM_DELETE_WINDOW {my exit}
	}
}
# End of main window ```1

include RosterItem TreeItem 0.95
resource icons {group away dnd ffc invisible message na occ offline online url}
MainWindow

